#!/usr/bin/env python
"""
IceProd Basic Submit
====================

Processes a script over jobs with input and output files.

Details
-------

Files are expected to be either full URLs or paths on the UW-Madison IceCube
file system.

Script arguments will be passed as a string. They can use these built-in macros:

* $(input) = The input file list, space-separated.
* $(output) = The output file.
* $(dataset) = The dataset_id in numerical form.
* $(job) = The job index within the dataset.

Example
-------

An example submission::

    ./basic_submit.py --token XXXX --env_shell '/cvmfs/icecube.opensciencegrid.org/py3-v4.1.1/icetray-env combo/stable' my_script.py '--foo=bar $(input) $(output)' job_files.txt

This will execute `my_script.py` from the local directory, while in the
combo/stable environment.  If the first line of `job_files.txt` contains::

    /data/user/XXX/gcdfile.i3.gz /data/user/XXX/infile_01.i3.gz /data/user/XXX/outfile_01.i3.gz

Then the first job will look like this::

    my_script.py --foo=bar gcdfile.i3.gz infile_01.i3.gz outfile_01.i3.gz

With data transfer of the script and files happening automatically.

Args:
    script (str): url or file path to script
    args (str): arguments to script
    files (str): filename with input and output files (one line per job)
    description: general description for dataset
    env_shell: an environment shell to execute before the script starts
    request_memory (float): request memory in GB (default: 1 GB)
    request_cpus (int): request CPUs (default: 1)
    request_gpus (int): request GPUs (default: 0)
    token (str): IceProd user token (https://iceprod2.icecube.wisc.edu/profile)
"""

import os
import sys
import argparse
import logging
import asyncio

from rest_tools.client import SavedDeviceGrantAuth


logger = logging.getLogger('basic_submit')

def fail(message):
    print(message)
    sys.exit(1)

async def run(rpc, rpc_materialization, args):
    def modify_path(src):
        if src.startswith('/cvmfs'):
            return src
        if os.path.exists(src):
            if not src.startswith('/data'):
                raise Exception('path must be in /data')
            return 'gsiftp://gridftp.icecube.wisc.edu' + os.path.abspath(src)
        elif not (src.startswith('http://') or src.startswith('https://') or src.startswith('gsiftp://')):
            raise Exception('unknown path: '+src)
        return src

    # check script location
    args['script'] = modify_path(args['script'])

    # check input and output files
    jobfiles = []
    with open(args['files']) as f:
        for line in f:
            files = [x.strip() for x in line.split() if x.strip()]
            if not files:
                continue
            outfiles = ['gsiftp://gridftp.icecube.wisc.edu' + os.path.abspath(files[-1])]
            jobfiles.append([modify_path(x) for x in files[:-1]]+outfiles)

    # make dataset config
    config = {
      "parent_id":0,
      "version":3,
      "options":{},
      "steering":{
        "parameters":{},
        "batchsys":None,
        "system":{},
        "resources":[],
        "data":[]
      },
      "tasks":[
        {
          "depends":[],
          "batchsys":None,
          "trays":[
            {
              "iterations":1,
              "modules":[
                {
                  "running_class":"",
                  "src":args['script'],
                  "args":args['args'],
                  "env_shell":args['env_shell'],
                  "env_clear":True,
                  "name":"",
                  "resources":[],
                  "data":[],
                  "classes":[],
                  "parameters":{},
                  "projects":[]
                }
              ],
              "name":"",
              "resources":[],
              "data":[],
              "classes":[],
              "parameters":{},
              "projects":[]
            }
          ],
          "requirements":{
            "memory": args['request_memory'],
            "cpu": args['request_cpus'],
            "gpu": args['request_gpus']
          },
          "name":"BasicSubmit",
          "task_files":True,
          "resources":[],
          "data":[],
          "classes":[],
          "parameters":{},
          "projects":[]
        }
      ],
      "difplus":None,
      "description":args['description'],
      "categories":[]
    }

    # create the dataset
    rpc_args = {
        'description': args['description'],
        'jobs_submitted': len(jobfiles),
        'tasks_submitted': len(jobfiles),
        'tasks_per_job': 1,
        'group': args['group'],
        'status': 'suspended',
    }
    try:
        ret = await rpc.request('POST', '/datasets', rpc_args)
    except Exception:
        logger.warning('failed to create dataset', exc_info=True)
        fail('Failed to create dataset')
    dataset_id = ret['result'].split('/')[-1]

    try:
        # upload config
        logger.info
        try:
            await rpc.request('PUT', f'/config/{dataset_id}', config)
        except Exception:
            logger.warning(f'uploading config failed for dataset_id {dataset_id}', exc_info=True)
            fail('Upload of dataset config failed')

        # materialize tasks
        try:
            ret = await rpc_materialization.request('POST', f'/request/{dataset_id}', {'num': len(jobfiles)})
        except Exception:
            logger.warning(f'materialization request for dataset {dataset_id} failed', exc_info=True)
            fail('Creation of jobs failed')
        materialization_id = ret['result']

        logger.info(f'waiting for materialization request')
        while True:
            await asyncio.sleep(10)
            ret = await rpc_materialization.request('GET', f'/request/{dataset_id}/status')
            if ret['status'] == 'complete':
                logger.info(f'materialization request complete')
                break
            elif ret['status'] == 'error':
                logger.warning(f'materialization request failed')
                fail('Creation of jobs failed')
            print('.', end='', flush=True)

        # upload file paths
        for i,files in enumerate(jobfiles):
            if i%100 == 0:
                print('.', end='', flush=True)
            try:
                for f in files[:-1]:
                    rpc_args = {
                        'filename': f,
                        'movement': 'input',
                        'job_index': i,
                        'task_name': 'BasicSubmit',
                    }
                    await rpc.request('POST', f'/datasets/{dataset_id}/files', rpc_args)
                rpc_args = {
                    'filename': files[-1],
                    'movement': 'output',
                    'job_index': i,
                    'task_name': 'BasicSubmit',
                }
                await rpc.request('POST', f'/datasets/{dataset_id}/files', rpc_args)
            except Exception:
                logger.warning(f'failed to upload file for job_index {i}', exc_info=True)
                fail('Failed to upload an input or output file path')

        # set dataset to processing
        rpc_args = {
            'status': 'processing'
        }
        try:
            await rpc.request('PUT', f'/datasets/{dataset_id}/status', rpc_args)
        except Exception:
            logger.warning(f'failed to set dataset {dataset_id} to processing', exc_info=True)
            fail('Failed to set dataset to processing')
    except Exception:
        rpc_args = {
            'status': 'failed'
        }
        try:
            await rpc.request('PUT', f'/datasets/{dataset_id}/status', rpc_args)
        except Exception:
            pass
        logger.warning(f'failed to start dataset', exc_info=True)
        fail('failed to start dataset')

    # get dataset num
    try:
        ret = await rpc.request('GET', f'/datasets/{dataset_id}', {'keys': 'dataset'})
        dataset_num = ret['dataset']
    except Exception:
        logger.warning(f'failed to get dataset info for dataset {dataset_id}', exc_info=True)
        fail('Failed to get final dataset info')

    print(f'Dataset {dataset_num} is running.')
    print(f'Check status at https://iceprod2.icecube.wisc.edu/dataset/{dataset_num}')

def main():
    parser = argparse.ArgumentParser(description='job completion')
    parser.add_argument('script', help='url or file path to script')
    parser.add_argument('args', help='arguments to script')
    parser.add_argument('files', help='filename with input and output files (one line per job)')
    parser.add_argument('--description', default='', help='general description for dataset')
    parser.add_argument('--group', default='users', choices=['users', 'simprod', 'filtering'],
                        help='group to run under (default: users)')
    parser.add_argument('--env_shell', default='', help='environment shell (icetray env_shell.sh syntax)')
    parser.add_argument('--request_memory', type=float, default=1.0,
                        help='request memory in GB (default: 1 GB)')
    parser.add_argument('--request_cpus', type=int, default=1,
                        help='request CPUs (default: 1)')
    parser.add_argument('--request_gpus', type=int, default=0,
                        help='request GPUs (default: 0)')
    parser.add_argument('--log_level', default='warning', help='log level (defaut: WARN)')

    args = parser.parse_args()
    args = vars(args)

    logging.basicConfig(level=getattr(logging, args['log_level'].upper()))

    rpc_kwargs = {
        'token_url': 'https://keycloak.icecube.wisc.edu/auth/realms/IceCube',
        'filename': os.path.abspath(os.path.expandvars('$HOME/.iceprod-auth')),
        'client_id': 'iceprod-public',
    }
    rpc = SavedDeviceGrantAuth('https://iceprod2-api.icecube.wisc.edu', **rpc_kwargs)
    rpc_materialization = SavedDeviceGrantAuth('https://materialization.iceprod.icecube.aq', **rpc_kwargs)

    asyncio.run(run(rpc, rpc_materialization, args))

if __name__ == '__main__':
    main()



"""
Mark datasets as completed.

Check jobs in a dataset, and when all are complete mark the dataset
as complete.
"""

import argparse
import asyncio
import logging

import requests.exceptions

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('dataset_completion')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    datasets = await rest_client.request('GET', '/dataset_summaries/status')
    dataset_ids = []
    if 'processing' in datasets:
        dataset_ids.extend(datasets['processing'])
    if 'truncated' in datasets:
        dataset_ids.extend(datasets['truncated'])
    for dataset_id in dataset_ids:
        # clean up misconfigured datasets that have no config
        try:
            try:
                await rest_client.request('GET', f'/config/{dataset_id}')
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404:
                    logger.info('dataset %s status -> suspended', dataset_id)
                    args = {'status': 'suspended'}
                    await rest_client.request('PUT', f'/datasets/{dataset_id}/status', args)
                    break
        except Exception:
            logger.error('error checking dataset %s config', dataset_id, exc_info=True)
            if debug:
                raise

        # test if dataset is complete / failed
        try:
            dataset = await rest_client.request('GET', f'/datasets/{dataset_id}')
            jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/job_counts/status')
            if sum(jobs[s] for s in jobs) >= dataset['jobs_submitted']:
                # jobs are all buffered / materialized
                job_statuses = set(jobs)
                if job_statuses == set(['complete']):
                    logger.info('dataset %s status -> complete', dataset_id)
                    args = {'status': 'complete'}
                    await rest_client.request('PUT', f'/datasets/{dataset_id}/status', args)
                elif 'processing' not in job_statuses:
                    if 'errors' in job_statuses:
                        logger.info('dataset %s status -> errors', dataset_id)
                        args = {'status': 'errors'}
                    else:
                        logger.info('dataset %s status -> suspended', dataset_id)
                        args = {'status': 'suspended'}
                    await rest_client.request('PUT', f'/datasets/{dataset_id}/status', args)
        except Exception:
            logger.error('error completing dataset %s', dataset_id, exc_info=True)
            if debug:
                raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Move tasks from reset to waiting/failed/suspended.
"""

import argparse
import asyncio
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('reset_tasks')


async def reset_dataset(dataset_id, rest_client=None, debug=False):
    try:
        dataset = await rest_client.request('GET', '/datasets/{}'.format(dataset_id))
        tasks = await rest_client.request('GET', '/datasets/{}/task_summaries/status'.format(dataset_id))
        if 'reset' in tasks:
            logger.info('dataset %s reset tasks: %s', dataset_id, tasks['reset'])
            for task_id in tasks['reset']:
                try:
                    task = await rest_client.request('GET', '/tasks/{}'.format(task_id))
                    status = 'waiting'
                    if dataset['debug']:
                        status = 'suspended'
                    elif 'failures' in task and task['failures'] > 10:
                        status = 'failed'
                    args = {'status': status}
                    await rest_client.request('PUT', '/datasets/{}/tasks/{}/status'.format(dataset_id,task_id), args)
                except Exception:
                    logger.error('error resetting task %s', task_id, exc_info=True)
                    if debug:
                        raise
    except Exception:
        logger.error('error resetting tasks in dataset %s', dataset_id, exc_info=True)
        if debug:
            raise


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    try:
        datasets = await rest_client.request('GET', '/dataset_summaries/status')
        dataset_ids = []
        if 'processing' in datasets:
            dataset_ids.extend(datasets['processing'])
        if 'truncated' in datasets:
            dataset_ids.extend(datasets['truncated'])
        awaitables = set()
        for dataset_id in dataset_ids:
            fut = asyncio.create_task(reset_dataset(dataset_id, rest_client=rest_client, debug=debug))
            awaitables.add(fut)
            if len(awaitables) >= 20:
                done, pending = await asyncio.wait(awaitables, return_when=asyncio.FIRST_COMPLETED)
                for fut in done:
                    await fut
                awaitables = pending
        for fut in asyncio.as_completed(awaitables):
            await fut
    except Exception:
        logger.error('error resetting non-active tasks', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--dataset-id', default=None, help='dataset id to reset')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    if args.dataset_id:
        asyncio.run(reset_dataset(args.dataset_id, rest_client, debug=args.debug))
    else:
        asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""Utility functions"""
from datetime import datetime


def datetime2str(dt):
    """Convert a datetime object to ISO 8601 string"""
    return dt.isoformat()


def nowstr():
    """Get an ISO 8601 string of the current time in UTC"""
    return datetime.utcnow().isoformat()


def str2datetime(st):
    """Convert a ISO 8601 string to datetime object"""
    if '.' in st:
        return datetime.strptime(st, "%Y-%m-%dT%H:%M:%S.%f")
    else:
        return datetime.strptime(st, "%Y-%m-%dT%H:%M:%S")


dataset_statuses = ['processing', 'truncated', 'suspended', 'errors', 'complete']
job_statuses = ['processing', 'suspended', 'errors', 'complete']
task_statuses = ['idle', 'waiting', 'queued', 'processing', 'reset',
                 'suspended', 'failed', 'complete']


def dataset_status_sort(st):
    try:
        return dataset_statuses.index(st)
    except ValueError:
        return len(dataset_statuses)


def job_status_sort(st):
    try:
        return job_statuses.index(st)
    except ValueError:
        return len(job_statuses)


def task_status_sort(st):
    try:
        return task_statuses.index(st)
    except ValueError:
        return len(task_statuses)




page_template = '''
    <h1>%s</h1>
    <p class = "desc">
    %s
    </p>
    <h2>Properties</h2>
    <ul>
    %s
    </ul>

    '''


# Converts ##LinkName## in string to html linkds
def add_links(desc):
    d = desc.split('##')
    is_link = False
    desc = ''
    for i in d:
        if is_link:
            desc += "<a href='javascript: show_doc(\"%s\");'>%s</a>" % (i, i)
        else:
            desc += i
        is_link = not is_link
    return desc


# Loads and parses documentation file
def load_doc(filename):
    try:
        f = open(filename.lower()+'.txt', 'rt')
    except IOError:
        return 'Not found'
    lines = []
    for line in f.readlines():
        if len(line.strip()):
            lines.append(line)

    body = ''
    for line in lines[1:]:
        nd = line.split(None, 1)
        name = nd[0]
        desc = nd[1] if len(nd) == 2 else ''
        desc = add_links(desc)
        body += '<li><h3>%s</h3><p>%s</p></li>' % (name, desc)

    nd = lines[0].split(None, 1)
    name = nd[0]
    desc = nd[1] if len(nd) == 2 else ''
    f.close()

    return page_template % (name, desc, body)



import asyncio
from concurrent.futures import ThreadPoolExecutor
import io
from functools import partial
import logging

try:
    import boto3
    import botocore.client
    import botocore.exceptions
except ImportError:
    boto3 = None

logger = logging.getLogger('s3')


class S3:
    """S3 wrapper for uploading and downloading objects"""
    def __init__(self, address, access_key, secret_key, bucket='iceprod2-logs'):
        self.s3 = None
        self.bucket = bucket
        try:
            self.s3 = boto3.client(
                's3',
                'us-east-1',
                endpoint_url=address,
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                config=botocore.client.Config(max_pool_connections=101)
            )
        except Exception:
            logger.warning('failed to connect to s3: %r', address, exc_info=True)
            raise
        self.executor = ThreadPoolExecutor(max_workers=20)

    async def get(self, key):
        """Download object from S3"""
        ret = ''
        with io.BytesIO() as f:
            loop = asyncio.get_running_loop()
            try:
                await loop.run_in_executor(self.executor, partial(self.s3.download_fileobj, Bucket=self.bucket, Key=key, Fileobj=f))
                ret = f.getvalue()
            except botocore.exceptions.ClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == '404':
                    return ''  # don't error on a 404
                raise
        return ret.decode('utf-8')

    def get_presigned(self, key, expiration=3600):
        """Make a presigned download url"""
        return self.s3.generate_presigned_url(
            'get_object',
            Params={'Bucket': self.bucket, 'Key': key},
            ExpiresIn=expiration,
        )

    async def put(self, key, data):
        """Upload object to S3"""
        with io.BytesIO(data.encode('utf-8')) as f:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(self.executor, partial(self.s3.upload_fileobj, f, self.bucket, key))

    def put_presigned(self, key, expiration=3600):
        """Make a presigned upload url"""
        return self.s3.generate_presigned_url(
            'put_object',
            Params={'Bucket': self.bucket, 'Key': key},
            ExpiresIn=expiration,
        )

    async def exists(self, key):
        """Check existence in S3"""
        loop = asyncio.get_running_loop()
        try:
            await loop.run_in_executor(self.executor, partial(self.s3.head_object, Bucket=self.bucket, Key=key))
        except Exception:
            return False
        return True

    async def delete(self, key):
        """Delete object in S3"""
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(self.executor, partial(self.s3.delete_object, Bucket=self.bucket, Key=key))


class FakeS3(S3):
    def __init__(self, s3override):
        self.s3 = s3override
        self.bucket = 'iceprod2-logs'
        self.executor = ThreadPoolExecutor(max_workers=20)



import logging

from rest_tools.client import ClientCredentialsAuth, SavedDeviceGrantAuth
from wipac_dev_tools import from_environment


def add_auth_to_argparse(parser):
    """Add auth args to argparse."""
    config = from_environment({
        'OAUTH_URL': 'https://keycloak.icecube.wisc.edu/auth/realms/IceCube',
        'OAUTH_CLIENT_ID': 'iceprod',
        'OAUTH_CLIENT_SECRET': '',
    })
    description = '''
        Use either user credentials or client credentials to authenticate.
        Can also be specified via env variables: OAUTH_URL, OAUTH_CLIENT_ID,
        and OAUTH_CLIENT_SECRET.
    '''
    parser.add_argument('--rest-url', default='https://iceprod2-api.icecube.wisc.edu',
                        help='URL for REST API (default: IceProd API)')
    group = parser.add_argument_group('OAuth', description)
    group.add_argument('--oauth-url', default=config['OAUTH_URL'],
                       help='The OAuth server URL for OpenID discovery')
    group.add_argument('--oauth-client-id', default=config['OAUTH_CLIENT_ID'],
                       help='The OAuth client id')
    group.add_argument('--oauth-client-secret', default=config['OAUTH_CLIENT_SECRET'],
                       help='The OAuth client secret, to enable client credential mode')


def create_rest_client(args):
    """Create a RestClient from argparse args."""
    if args.oauth_client_secret:
        logging.debug('Using client credentials to authenticate')
        return ClientCredentialsAuth(
            address=args.rest_url,
            token_url=args.oauth_url,
            client_id=args.oauth_client_id,
            client_secret=args.oauth_client_secret,
        )
    else:
        logging.debug('Using user credentials to authenticate')
        if args.oauth_client_id == 'iceprod':
            args.oauth_client_id = 'iceprod-public'
        return SavedDeviceGrantAuth(
            address=args.rest_url,
            filename='.iceprod-auth',
            token_url=args.oauth_url,
            client_id=args.oauth_client_id,
        )



"""
Mark jobs as completed.

Check tasks in a job, and when all are complete mark the job as complete.
"""

import argparse
import asyncio
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('job_completion')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    datasets = await rest_client.request('GET', '/dataset_summaries/status')
    dataset_ids = []
    if 'processing' in datasets:
        dataset_ids.extend(list(datasets['processing']))
    if 'truncated' in datasets:
        dataset_ids.extend(list(datasets['truncated']))
    for dataset_id in dataset_ids:
        try:
            jobs = await rest_client.request('GET', '/datasets/{}/job_summaries/status'.format(dataset_id))
            if 'processing' not in jobs:
                continue
            for job_id in jobs['processing']:
                tasks = await rest_client.request('GET', '/datasets/{}/tasks?keys=task_id|task_index|status&job_id={}'.format(dataset_id,job_id))
                task_statuses = []
                for task_id in sorted(tasks, key=lambda t:tasks[t]['task_index']):
                    task_statuses.append(tasks[task_id]['status'])
                if set(task_statuses) == set(['complete']):
                    logger.info('dataset %s job %s status -> complete', dataset_id, job_id)
                    args = {'status':'complete'}
                    await rest_client.request('PUT', '/datasets/{}/jobs/{}/status'.format(dataset_id,job_id), args)
                for status in task_statuses:
                    if status not in ('idle','waiting','queued','processing','reset','complete'):
                        if status == 'failed':
                            logger.info('dataset %s job %s status -> errors', dataset_id, job_id)
                            args = {'status':'errors'}
                        else:
                            logger.info('dataset %s job %s status -> suspended', dataset_id, job_id)
                            args = {'status':'suspended'}
                        await rest_client.request('PUT', '/datasets/{}/jobs/{}/status'.format(dataset_id,job_id), args)
                        break
        except Exception:
            logger.error('error completing a job in dataset %s', dataset_id, exc_info=True)
            if debug:
                raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""
The task runner.

Run it with `python -m iceprod.core.i3exec`.

optional arguments:
  -h, --help            show this help message and exit
  -f CFGFILE, --cfgfile CFGFILE
                        Specify config file
  -u URL, --url URL     URL of the iceprod server
  -p PASSKEY, --passkey PASSKEY
                        passkey for communcation with iceprod server
  --pilot_id PILOTID    ID of the pilot (if this is a pilot)
  -d, --debug           Enable debug actions and logging
  --offline             Enable offline mode (don't talk with server)
  --offline_transfer True/False
                        Enable offline file transfer
  --logfile LOGFILE     Specify the logfile to use
  --job JOB             Index of the job to run
  --task TASK           Name of the task to run
  --gzip-logs           gzip the iceprod logs
"""

from __future__ import absolute_import, division, print_function

import os
import logging
import logging.config
import time
import socket
from functools import partial
import shutil
import threading
import asyncio
import gzip

import iceprod
from iceprod.core import constants
import iceprod.core.dataclasses
import iceprod.core.serialization
from iceprod.core.defaults import add_default_options
import iceprod.core.exe
from iceprod.core.exe_json import ServerComms
import iceprod.core.pilot
import iceprod.core.resources
import iceprod.core.logger


def load_config(cfgfile):
    """Load a config from file, serialized string, dictionary, etc"""
    logger = logging.getLogger('i3exec')
    config = None
    if isinstance(cfgfile,str):
        try:
            if os.path.exists(cfgfile):
                config = iceprod.core.serialization.serialize_json.load(cfgfile)
            else:
                config = iceprod.core.serialization.serialize_json.loads(cfgfile)
            if not config:
                raise Exception('Config not found')
        except Exception as e:
            logger.critical('Error loading configuration: %s' % str(e))
            raise
    elif isinstance(cfgfile,iceprod.core.dataclasses.Job):
        config = cfgfile
    elif isinstance(cfgfile,dict):
        config = iceprod.core.serialization.dict_to_dataclasses(cfgfile)
    else:
        logger.warning('cfgfile: %r',cfgfile)
        raise Exception('cfgfile is not a str or a Job')
    return config


def main(cfgfile=None, logfile=None, url=None, debug=False,
         passkey='', pilot_id=None, offline=False, offline_transfer=False,
         gzip_logs=False):
    """Main task runner for iceprod"""
    # set up logger
    if debug:
        logl = 'INFO'
    else:
        logl = 'WARNING'
    if logfile:
        logf = os.path.abspath(os.path.expandvars(logfile))
    else:
        logf = os.path.abspath(os.path.expandvars(constants['stdlog']))
        if gzip_logs:
            logf += '.gz'
    iceprod.core.logger.set_logger(loglevel=logl,
                                   logfile=logf,
                                   logsize=67108864,
                                   lognum=1)
    logging.warning('starting IceProd core')

    if cfgfile is None:
        logging.critical('There is no cfgfile')
        raise Exception('missing cfgfile')
    elif isinstance(cfgfile, str):
        config = load_config(cfgfile)
    else:
        config = cfgfile
    logging.info('config: %r',config)

    if not offline:
        # if we are not in offline mode, we need a url
        if not url:
            logging.critical('url missing')
            raise Exception('url missing')

        # setup server comms
        kwargs = {}
        if 'username' in config['options']:
            kwargs['username'] = config['options']['username']
        if 'password' in config['options']:
            kwargs['password'] = config['options']['password']
        if 'ssl' in config['options'] and config['options']['ssl']:
            kwargs.update(config['options']['ssl'])
        rpc = ServerComms(url, passkey, None, **kwargs)

    async def run():
        if offline:
            logging.info('offline mode')
            async for proc in runner(config, debug=debug, offline=offline,
                                     offline_transfer=offline_transfer):
                await proc.wait()
        elif 'tasks' in config and config['tasks']:
            logging.info('online mode - single task')
            # tell the server that we are processing this task
            if 'task_id' not in config['options']:
                raise Exception('config["options"]["task_id"] not specified')
            pilot_id2 = None
            try:
                pilot_id2 = await rpc.create_pilot(queue_host='manual',
                                                   host=socket.getfqdn(),
                                                   grid_queue_id='1',
                                                   queue_version=iceprod.__version__,
                                                   version=iceprod.__version__,
                                                   tasks=[config['options']['task_id']],
                                                   resources={})
                await rpc.processing(config['options']['task_id'])
            except Exception:
                logging.error('comms error', exc_info=True)

            # set up stdout and stderr
            try:
                async for proc in runner(config, rpc=rpc, debug=debug):
                    await proc.wait()
            finally:
                try:
                    if pilot_id2:
                        await rpc.delete_pilot(pilot_id2)
                except Exception:
                    logging.error('comms error', exc_info=True)
        else:
            logging.info('pilot mode - get many tasks from server')
            if 'gridspec' not in config['options']:
                logging.critical('gridspec missing')
                raise Exception('gridspec missing')
            if not pilot_id:
                logging.critical('pilot_id missing')
                raise Exception('pilot_id missing')
            pilot_kwargs = {}
            if 'run_timeout' in config['options']:
                pilot_kwargs['run_timeout'] = config['options']['run_timeout']
            if 'restrict_site' in config['options']:
                pilot_kwargs['restrict_site'] = config['options']['restrict_site']
            async with iceprod.core.pilot.Pilot(
                    config, rpc=rpc, debug=debug,
                    runner=partial(runner, rpc=rpc, debug=debug),
                    pilot_id=pilot_id, **pilot_kwargs) as p:
                await p.run()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run())

    logging.warning('finished running normally; exiting...')
    iceprod.core.logger.remove_handlers()


async def runner(config, rpc=None, debug=False, offline=False,
                 offline_transfer=False, resources=None):
    """Run a config.

    #. Set some default options if not set in configuration.
    #. Set up global env based on the configuration.
    #. Run tasks
       * If a task is specified in the configuration options:

         If the task is specified by name or number, run only that task.
         If there is a problem finding the task specified, raise a
         critical error.

       * Otherwise, run all tasks in the configuration in the order
         they were written.

    #. Destroy the global env, uploading and deleting files as needed.
    #. Upload the log, error, and output files if specified in options.

    Args:
        config (`iceprod.core.dataclasses.Job`): Dataset configuration
        rpc (:py:class:`iceprod.core.exe_json.ServerComms`): RPC object
        debug (bool): (optional) turn on debug logging
        offline (bool): (optional) enable offline mode
        offline_transfer (bool): (optional) enable/disable offline data transfers
        resources (:py:class:`iceprod.core.resources.Resources`): (optional) Resources object
    """
    # set logging
    if offline:
        logger = logging.getLogger('task')
        if 'task_id' not in config['options']:
            config['options']['task_id'] = 'a'
    else:
        if 'task_id' not in config['options']:
            raise Exception('task_id not set in config options')
        logger = logging.getLogger(config['options']['task_id'])
    if 'debug' not in config['options']:
        config['options']['debug'] = debug
    if ('debug' in config['options'] and config['options']['debug'] and
            'loglevel' not in config['options']):
        config['options']['loglevel'] = 'INFO'
    if ('loglevel' in config['options'] and
            config['options']['loglevel'].upper() in iceprod.core.logger.setlevel):
        try:
            iceprod.core.logger.set_log_level(config['options']['loglevel'])
        except Exception:
            logger.warning('failed to set a new log level', exc_info=True)

    # make sure some basic options are set
    add_default_options(config['options'])
    if 'offline' not in config['options']:
        config['options']['offline'] = offline
    if 'offline_transfer' not in config['options']:
        config['options']['offline_transfer'] = offline_transfer
    if 'subprocess_dir' not in config['options']:
        config['options']['subprocess_dir'] = os.getcwd()
    if 'task_temp' not in config['options']:
        config['options']['task_temp'] = 'file:'+os.path.join(config['options']['subprocess_dir'],'task_temp')
    if 'tray_temp' not in config['options']:
        config['options']['tray_temp'] = 'file:'+os.path.join(config['options']['subprocess_dir'],'tray_temp')
    if 'local_temp' not in config['options']:
        config['options']['local_temp'] = os.path.join(config['options']['subprocess_dir'],'local_temp')
    if 'credentials_dir' not in config['options']:
        config['options']['credentials_dir'] = os.path.join(os.getcwd(), 'iceprod_credentials')
    if 'stillrunninginterval' not in config['options']:
        config['options']['stillrunninginterval'] = 60
    if 'upload' not in config['options']:
        config['options']['upload'] = 'logging'

    if not config['steering']:
        # make sure steering exists in the config
        config['steering'] = iceprod.core.dataclasses.Steering()

    resource_thread = None
    if not resources:
        try:
            import psutil
        except ImportError:
            resources = None
        else:
            # track resource usage in separate thread
            resource_stop = False
            resources = iceprod.core.resources.Resources(debug=debug)
            config['options']['resources'] = resources.claim(config['options']['task_id'])
            resources.register_process(config['options']['task_id'],
                                       psutil.Process(), os.getcwd())

            def track():
                while not resource_stop:
                    resources.check_claims()
                    time.sleep(1)
            resource_thread = threading.Thread(target=track)
            resource_thread.start()

    # make exe Config
    cfg = iceprod.core.exe.Config(config=config, rpc=rpc, logger=logger)

    # set up global env, based on config['options'] and config.steering
    env_opts = cfg.parseObject(config['options'], {})
    stats = {}
    try:
        try:
            # keep track of the start time
            start_time = time.time()
            stats = {}

            async with iceprod.core.exe.SetupEnv(cfg, config['steering'], {'options':env_opts}, logger=logger) as env:
                logger.warning("config options: %r",config['options'])

                # find tasks to run
                if 'task' in config['options']:
                    logger.warning('task specified: %r',config['options']['task'])
                    # run only this task name or number
                    name = config['options']['task']
                    if isinstance(name, iceprod.core.dataclasses.String) and name.isdigit():
                        name = int(name)
                    if isinstance(name, iceprod.core.dataclasses.String):
                        # find task by name
                        for task in config['tasks']:
                            if task['name'] == name:
                                async for proc in iceprod.core.exe.runtask(cfg, env, task, logger=logger):
                                    yield proc
                                break
                        else:
                            logger.critical('cannot find task named %r', name)
                            raise Exception('cannot find specified task')
                    elif isinstance(name, int):
                        # find task by index
                        if name >= 0 and name < len(config['tasks']):
                            async for proc in iceprod.core.exe.runtask(cfg, env, config['tasks'][name], logger=logger):
                                yield proc
                        else:
                            logger.critical('cannot find task index %d', name)
                            raise Exception('cannot find specified task')

                    else:
                        logger.critical('task specified in options is %r, but no task found',
                                        name)
                        raise Exception('cannot find specified task')
                    stats = env['stats']
                elif offline:
                    # run all tasks in order
                    for task in config['tasks']:
                        async for proc in iceprod.core.exe.runtask(cfg, env, task):
                            yield proc
                else:
                    raise Exception('task to run not specified')

            if (not offline) and 'task' in config['options']:
                # finish task
                logger.warning('task finishing')
                await rpc.finish_task(
                    config['options']['task_id'],
                    dataset_id=config['options']['dataset_id'],
                    stats=stats, start_time=start_time,
                    resources=resources.get_final(config['options']['task_id']) if resources else None,
                    site=resources.site if resources else None,
                )

        except Exception as e:
            logger.error('task failed, exiting without running completion steps.',
                         exc_info=True)
            # set task status on server
            if not offline:
                try:
                    await rpc.task_error(
                        config['options']['task_id'],
                        dataset_id=config['options']['dataset_id'],
                        stats=stats, start_time=start_time,
                        reason=str(e),
                        resources=resources.get_final(config['options']['task_id']) if resources else None,
                        site=resources.site if resources else None,
                    )
                except Exception as e:
                    logger.error(e)
                # forcibly turn on logging, so we can see the error
                config['options']['upload'] = 'logging'
            raise

    finally:
        # check resources
        if resource_thread:
            resource_stop = True
            resource_thread.join()
            print('Resources:')
            r = resources.get_final(config['options']['task_id'])
            if not r:
                print('  None')
            else:
                for k in r:
                    print('  {}: {:.2f}'.format(k,r[k]))
        # upload log files to server
        try:
            if (not offline) and 'upload' in config['options']:
                logger.warning('uploading logfiles')
                if isinstance(config['options']['upload'],
                              iceprod.core.dataclasses.String):
                    upload = config['options']['upload'].lower().split('|')
                elif isinstance(config['options']['upload'],(tuple,list)):
                    upload = [x.lower() for x in config['options']['upload']]
                else:
                    raise Exception('upload config is not a valid type')
                errfile = constants['stderr']
                outfile = constants['stdout']
                if 'subprocess_dir' in config['options'] and config['options']['subprocess_dir']:
                    subdir = config['options']['subprocess_dir']
                    errfile = os.path.join(subdir, errfile)
                    outfile = os.path.join(subdir, outfile)
                for up in upload:
                    if up.startswith('logging'):
                        # upload err,log,out files
                        await rpc.uploadLog(task_id=config['options']['task_id'],
                                            dataset_id=config['options']['dataset_id'])
                        await rpc.uploadErr(filename=errfile,
                                            task_id=config['options']['task_id'],
                                            dataset_id=config['options']['dataset_id'])
                        await rpc.uploadOut(filename=outfile,
                                            task_id=config['options']['task_id'],
                                            dataset_id=config['options']['dataset_id'])
                        break
                    elif up.startswith('log'):
                        # upload log files
                        await rpc.uploadLog(task_id=config['options']['task_id'],
                                            dataset_id=config['options']['dataset_id'])
                    elif up.startswith('err'):
                        # upload err files
                        await rpc.uploadErr(filename=errfile,
                                            task_id=config['options']['task_id'],
                                            dataset_id=config['options']['dataset_id'])
                    elif up.startswith('out'):
                        # upload out files
                        await rpc.uploadOut(filename=outfile,
                                            task_id=config['options']['task_id'],
                                            dataset_id=config['options']['dataset_id'])
        except Exception:
            logger.error('failed when uploading logging info',exc_info=True)

    logger.warning('finished without error')


if __name__ == '__main__':
    # get arguments
    import argparse
    parser = argparse.ArgumentParser(description='IceProd Core')
    parser.add_argument('-f','--cfgfile', type=str,
                        help='Specify config file')
    parser.add_argument('-u','--url', type=str,
                        help='URL of the iceprod server')
    parser.add_argument('-p','--passkey', type=str,
                        help='passkey for communcation with iceprod server')
    parser.add_argument('--pilot_id', type=str, default=None,
                        help='ID of the pilot (if this is a pilot)')
    parser.add_argument('-d','--debug', action='store_true', default=False,
                        help='Enable debug actions and logging')
    parser.add_argument('--offline', action='store_true', default=False,
                        help='Enable offline mode (don\'t talk with server)')
    parser.add_argument('--offline_transfer', type=bool, default=False,
                        help='Enable/disable file transfer during offline mode')
    parser.add_argument('--logfile', type=str, default=None,
                        help='Specify the logfile to use')
    parser.add_argument('--dataset', type=int, default=None,
                        help='Dataset number')
    parser.add_argument('--job', type=int, default=None,
                        help='Index of the job to run')
    parser.add_argument('--jobs_submitted', type=int, default=None,
                        help='Total number of jobs in this dataset')
    parser.add_argument('--task', type=str, default=None,
                        help='Name of the task to run')
    parser.add_argument('--gzip-logs', dest='gzip_logs', action='store_true',
                        help='gzip the iceprod logs')

    args = vars(parser.parse_args())
    print(args)

    # check cfgfile
    if args['cfgfile'] is not None and not os.path.isfile(args['cfgfile']):
        if os.path.isfile(os.path.join(os.getcwd(),args['cfgfile'])):
            args['cfgfile'] = os.path.join(os.getcwd(),args['cfgfile'])
        else:
            args['cfgfile'] = None

    options = {k: args.pop(k) for k in ('dataset','job','jobs_submitted','task')}
    if not options['jobs_submitted'] and options['job']:
        options['jobs_submitted'] = options['job']+1
    options['debug'] = args['debug']
    if args['cfgfile']:
        cfgfile = load_config(args['cfgfile'])
        for k in options:
            if options[k] is not None and k not in cfgfile['options']:
                cfgfile['options'][k] = options[k]
        args['cfgfile'] = cfgfile

    # start iceprod
    try:
        main(**args)
    finally:
        if args['gzip_logs']:
            # compress stdout and stderr
            for filename in ('stdout', 'stderr'):
                if os.path.exists(constants[filename]):
                    with open(constants[filename], 'rb') as f_in:
                        with gzip.open(constants[filename]+'.gz', 'wb') as f_out:
                            shutil.copyfileobj(f_in, f_out)
                else:
                    with gzip.open(constants[filename]+'.gz', 'wb') as f_out:
                        pass



"""
A supercomputer plugin, specificially for
`Graham <https://docs.computecanada.ca/wiki/Graham>`_.
"""
import os
import stat
import logging
import getpass
import socket
from datetime import datetime,timedelta
import subprocess
import asyncio
from functools import partial

from tornado.concurrent import run_on_executor

import iceprod
from iceprod.core import constants
from iceprod.core.exe_json import ServerComms
from iceprod.server import grid
from iceprod.server.globus import SiteGlobusProxy

logger = logging.getLogger('plugin-graham')


async def subprocess_ssh(host, args):
    """A subprocess call over ssh"""
    cmd = ['ssh', '-i', '~/.ssh/iceprod', host]
    if 'ICEPRODBASE' in os.environ:
        cmd.append(f'{os.environ["ICEPRODBASE"]}/env-shell.sh')
    cmd += args
    logger.info('subprocess_ssh: %r', cmd)
    p = await asyncio.create_subprocess_exec(*cmd)
    await asyncio.wait_for(p.wait(), timeout=20*60)
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return p


async def check_call(*args, **kwargs):
    logger.info('subprocess_check_call: %r', args)
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return p


async def check_output(*args, **kwargs):
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.STDOUT
    logger.info('subprocess_check_output: %r', args)
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    out,_ = await p.communicate()
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return out.decode('utf-8')


class MyServerComms(ServerComms):
    def __init__(self, rest_client):
        self.rest = rest_client


class supercomp_graham(grid.BaseGrid):
    """Plugin Overrides for Graham SuperComputer"""

    # graham gpu queue requirements
    resources = {
        'cpu': 2,
        'memory': 8.,
        'disk': 48.,
        'gpu': 1,
        'time': 4.,
    }

    def __init__(self, *args, **kwargs):
        super(supercomp_graham, self).__init__(*args, **kwargs)
        self.x509proxy = SiteGlobusProxy()

    async def upload_logfiles(self, task_id, dataset_id, submit_dir='', reason=''):
        """upload logfiles"""
        data = {'name': 'stdlog', 'task_id': task_id, 'dataset_id': dataset_id}

        # upload stdlog
        filename = os.path.join(submit_dir, constants['stdlog'])
        if os.path.exists(filename):
            with open(filename) as f:
                data['data'] = f.read()
            await self.rest_client.request('POST', '/logs', data)
        else:
            data['data'] = reason
            await self.rest_client.request('POST', '/logs', data)

        # upload stderr
        data['name'] = 'stderr'
        filename = os.path.join(submit_dir, constants['stderr'])
        if os.path.exists(filename):
            with open(filename) as f:
                data['data'] = f.read()
            await self.rest_client.request('POST', '/logs', data)
        else:
            data['data'] = ''
            await self.rest_client.request('POST', '/logs', data)

        # upload stdout
        data['name'] = 'stdout'
        filename = os.path.join(submit_dir, constants['stdout'])
        if os.path.exists(filename):
            with open(filename) as f:
                data['data'] = f.read()
            await self.rest_client.request('POST', '/logs', data)
        else:
            data['data'] = ''
            await self.rest_client.request('POST', '/logs', data)

    async def task_error(self, task_id, dataset_id, submit_dir, reason=''):
        """reset a task"""
        # search for resources in slurm stdout
        resources = {}
        filename = os.path.join(submit_dir, 'slurm.out')
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break
        if not reason:
            # search for reason in logfile
            filename = os.path.join(submit_dir, constants['stdlog'])
            if os.path.exists(filename):
                with open(filename) as f:
                    for line in f:
                        line = line.strip()
                        if 'failed to download' in line:
                            reason = 'failed to download input file(s)'
                        if 'failed to upload' in line:
                            reason = 'failed to upload output file(s)'
                        if 'Exception' in line and not reason:
                            reason = line
                        if 'return code:' in line and not reason:
                            reason = 'task error: return code '+line.rsplit(':',1)[-1]

        comms = MyServerComms(self.rest_client)
        await comms.task_kill(task_id, dataset_id=dataset_id,
                              reason=reason, resources=resources)

    async def finish_task(self, task_id, dataset_id, submit_dir):
        """complete a task"""
        # search for reasources in slurm stdout
        resources = {}
        filename = os.path.join(submit_dir, 'slurm.out')
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break

        comms = MyServerComms(self.rest_client)
        await comms.finish_task(task_id, dataset_id=dataset_id,
                                resources=resources)

    async def check_and_clean(self):
        """Check and clean the grid"""
        host = socket.getfqdn()
        self.x509proxy.update_proxy()

        # first, check if anything has completed successfully
        grid_jobs = await self.get_grid_completions()

        if grid_jobs:
            # get pilots from iceprod
            pilots = await self.rest_client.request('GET', '/pilots')
            pilot_futures = []
            pilots_to_delete = set()
            for pilot_id in pilots:
                pilot = pilots[pilot_id]
                if pilot['queue_host'] != host:
                    continue
                if 'grid_queue_id' in pilot and pilot['grid_queue_id'] in grid_jobs:
                    try:
                        gid = pilot['grid_queue_id']
                        pilot['submit_dir'] = grid_jobs[gid]['submit_dir']
                        if pilot['tasks']:
                            task_id = pilot['tasks'][0]
                            ret = await self.rest_client.request('GET', f'/tasks/{task_id}')
                            if ret['status'] == 'processing':
                                pilot['dataset_id'] = ret['dataset_id']
                                if grid_jobs[gid]['status'] == 'ok':
                                    # upload data - do other steps in next loop
                                    pilot_futures.append(asyncio.ensure_future(self.upload_output(pilot)))
                                else:
                                    await self.upload_logfiles(task_id, pilot['dataset_id'],
                                                               submit_dir=pilot['submit_dir'])
                                    await self.task_error(task_id, pilot['dataset_id'],
                                                          submit_dir=pilot['submit_dir'])
                    except Exception:
                        logger.error('error handling task', exc_info=True)

                    pilots_to_delete.add(pilot_id)

            for fut in asyncio.as_completed(pilot_futures):
                pilot,e = await fut  # upload is done

                try:
                    task_id = pilot['tasks'][0]
                    if e:
                        reason = f'failed to upload output files\n{e}'
                        await self.upload_logfiles(task_id,
                                                   dataset_id=pilot['dataset_id'],
                                                   submit_dir=pilot['submit_dir'],
                                                   reason=reason)
                        await self.task_error(task_id,
                                              dataset_id=pilot['dataset_id'],
                                              submit_dir=pilot['submit_dir'],
                                              reason=reason)
                    else:
                        await self.upload_logfiles(task_id,
                                                   dataset_id=pilot['dataset_id'],
                                                   submit_dir=pilot['submit_dir'])
                        await self.finish_task(task_id,
                                               dataset_id=pilot['dataset_id'],
                                               submit_dir=pilot['submit_dir'])
                except Exception:
                    logger.error('error handling task', exc_info=True)

            for pilot_id in pilots_to_delete:
                await self.rest_client.request('DELETE', f'/pilots/{pilot_id}')

        # now, do regular check_and_clean
        await super(supercomp_graham, self).check_and_clean()

    async def queue(self):
        """Submit a pilot for each task, up to the limit"""
        host = socket.getfqdn()
        self.x509proxy.update_proxy()
        resources = self.resources.copy()

        dataset_cache = {}
        task_futures = []
        for _ in range(self.get_queue_num()):
            # get a processing task
            args = {'requirements': supercomp_graham.resources.copy()}
            args['requirements']['os'] = 'RHEL_7_x86_64'
            try:
                task = await self.rest_client.request('POST', '/task_actions/process', args)
            except Exception:
                logger.info('no more tasks to queue')
                break

            # get full job, dataset, config info
            job = await self.rest_client.request('GET', f'/jobs/{task["job_id"]}')
            if task['dataset_id'] in dataset_cache:
                dataset, config = dataset_cache[task['dataset_id']]
            else:
                dataset = await self.rest_client.request('GET', f'/datasets/{task["dataset_id"]}')
                config = await self.rest_client.request('GET', f'/config/{task["dataset_id"]}')
                dataset_cache[task['dataset_id']] = (dataset, config)

            # check if we have any files in the task_files API
            task_cfg = None
            for t in config['tasks']:
                if t['name'] == task['name']:
                    task_cfg = t
                    break
            else:
                logger.warning('cannot find task in config for %s', task['task_id'])
                continue
            if 'task_files' in task_cfg and task_cfg['task_files']:
                comms = MyServerComms(self.rest_client)
                files = await comms.task_files(task['dataset_id'],
                                               task['task_id'])
                task_cfg['data'].extend(files)

            config['dataset'] = dataset['dataset']
            task.update({
                'config': config,
                'job': job['job_index'],
                'dataset': dataset['dataset'],
                'jobs_submitted': dataset['jobs_submitted'],
                'tasks_submitted': dataset['tasks_submitted'],
                'debug': dataset['debug'],
                'requirements': args['requirements'],
            })

            # setup submit dir
            await self.setup_submit_directory(task)

            # create pilot
            resources_available = {'time': resources['time']}
            for k in ('cpu','gpu','memory','disk'):
                resources_available[k] = resources[k]-task['requirements'][k]
            pilot = {
                'resources': resources,
                'resources_available': resources_available,
                'resources_claimed': task['requirements'],
                'tasks': [task['task_id']],
                'queue_host': host,
                'queue_version': iceprod.__version__,
                'version': iceprod.__version__,
            }
            ret = await self.rest_client.request('POST', '/pilots', pilot)
            pilot['pilot_id'] = ret['result']
            task['pilot'] = pilot

            # get input files, all tasks in parallel
            task_futures.append(asyncio.ensure_future(self.download_input(task)))

        # wait for the futures
        for fut in asyncio.as_completed(task_futures):
            task,e = await fut
            try:
                pilot_id = task['pilot']['pilot_id']
                if e:
                    reason = f'failed to download input files\n{e}'
                    await self.upload_logfiles(task['task_id'],
                                               dataset_id=task['dataset_id'],
                                               submit_dir=task['submit_dir'],
                                               reason=reason)
                    await self.task_error(task['task_id'],
                                          dataset_id=task['dataset_id'],
                                          submit_dir=task['submit_dir'],
                                          reason=reason)
                    await self.rest_client.request('DELETE', f'/pilots/{pilot_id}')
                    continue

                # submit to queue
                await self.submit(task)

                # update pilot
                args = {'grid_queue_id': task['grid_queue_id']}
                await self.rest_client.request('PATCH', f'/pilots/{pilot_id}', args)
            except Exception as e:
                try:
                    reason = f'failed to submit pilot:\n{e}'
                    await self.upload_logfiles(task['task_id'],
                                               dataset_id=task['dataset_id'],
                                               submit_dir=task['submit_dir'],
                                               reason=reason)
                    await self.task_error(task['task_id'],
                                          dataset_id=task['dataset_id'],
                                          submit_dir=task['submit_dir'],
                                          reason=reason)
                except Exception:
                    pass
                logger.error('error handling pilot', exc_info=True)

    async def download_input(self, task):
        """
        Download input files for task.

        Args:
            task (dict): task info
        Returns:
            dict: task info
        """
        try:
            # proxy = self.x509proxy.get_proxy()
            await subprocess_ssh(
                'gra-dtn1.computecanada.ca',
                ['python', '-m', 'iceprod.core.data_transfer', '-f',
                 os.path.join(task['submit_dir'],'task.cfg'),
                 '-d', task['submit_dir'],
                 'input']
            )
        except Exception as e:
            return (task, e)
        return (task,None)

    async def upload_output(self, task):
        """
        Upload output files for task.

        Args:
            task (dict): task info
        Returns:
            dict: task info
        """
        try:
            # proxy = self.x509proxy.get_proxy()
            await subprocess_ssh(
                'gra-dtn1.computecanada.ca',
                ['python', '-m', 'iceprod.core.data_transfer', '-f',
                 os.path.join(task['submit_dir'],'task.cfg'),
                 '-d', task['submit_dir'],
                 'output']
            )
        except Exception as e:
            return (task,e)
        return (task,None)

    @run_on_executor
    def generate_submit_file(self, task, cfg=None, passkey=None,
                             filelist=None):
        """Generate queueing system submit file for task in dir."""
        args = self.get_submit_args(task,cfg=cfg)
        args.append('--offline')

        # write the submit file
        submit_file = os.path.join(task['submit_dir'],'submit.sh')
        with open(submit_file,'w') as f:
            p = partial(print,sep='',file=f)
            p('#!/bin/bash')
            p('#SBATCH --account=def-dgrant')
            p('#SBATCH --output={}'.format(os.path.join(task['submit_dir'],'slurm.out')))
            p('#SBATCH --error={}'.format(os.path.join(task['submit_dir'],'slurm.err')))
            p(f'#SBATCH --chdir={task["submit_dir"]}')
            p('#SBATCH --ntasks=1')
            p('#SBATCH --export=NONE')
            p('#SBATCH --mail-type=NONE')
            p('#SBATCH --job-name=iceprod_{}'.format(os.path.basename(task['submit_dir'])))

            # handle resource requests
            if 'requirements' in task:
                if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                    p(f'#SBATCH --cpus-per-task={task["requirements"]["cpu"]}')
                if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                    p(f'#SBATCH --gres=gpu:{task["requirements"]["gpu"]}')
                if 'memory' in task['requirements'] and task['requirements']['memory']:
                    p('#SBATCH --mem={}M'.format(int(task['requirements']['memory']*1000)))
                # we don't currently use the local disk, just the global scratch
                # if 'disk' in task['requirements'] and task['requirements']['disk']:
                #     p('#SBATCH --tmp={}M'.format(int(task['requirements']['disk']*1000)))
                if 'time' in task['requirements'] and task['requirements']['time']:
                    p('#SBATCH --time={}'.format(int(task['requirements']['time']*60)))

            # get batchopts
            for b in self.queue_cfg['batchopts']:
                p(b+'='+self.queue_cfg['batchopts'][b])

            # make resources explicit in env
            if 'requirements' in task:
                if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                    p(f'export NUM_CPUS={task["requirements"]["cpu"]}')
                if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                    p(f'export NUM_GPUS={task["requirements"]["gpu"]}')
                if 'memory' in task['requirements'] and task['requirements']['memory']:
                    p(f'export NUM_MEMORY={task["requirements"]["memory"]}')
                if 'disk' in task['requirements'] and task['requirements']['disk']:
                    p(f'export NUM_DISK={task["requirements"]["disk"]}')
                if 'time' in task['requirements'] and task['requirements']['time']:
                    p(f'export NUM_TIME={task["requirements"]["time"]}')

            p('{} {}'.format(os.path.join(task['submit_dir'],'loader.sh'), ' '.join(args)))

        # make it executable
        st = os.stat(submit_file)
        os.chmod(submit_file, st.st_mode | stat.S_IEXEC)

    async def submit(self,task):
        """Submit task to queueing system."""
        cmd = ['sbatch','submit.sh']
        ret = await check_output(*cmd, cwd=task['submit_dir'])
        grid_queue_id = ''
        for line in ret.split('\n'):
            if 'Submitted batch job' in line:
                grid_queue_id = line.strip().rsplit(' ',1)[-1]
                break
        else:
            raise Exception('did not get a grid_queue_id')
        task['grid_queue_id'] = grid_queue_id

    async def get_grid_status(self):
        """
        Get all tasks running on the queue system.

        Returns:
            dict: {grid_queue_id: {status, submit_dir} }
        """
        cmd = ['squeue', '-u', getpass.getuser(), '-h', '-o', '%A %t %j %o']
        out = await check_output(*cmd)
        ret = {}
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,name,cmd = line.split()
            if not name.startswith('iceprod'):
                continue
            if status == 'PD':
                status = 'queued'
            elif status == 'R':
                status = 'processing'
            elif status == 'CD':
                status = 'completed'
            else:
                status = 'error'
            ret[gid] = {'status':status,'submit_dir':os.path.dirname(cmd)}
        return ret

    async def get_grid_completions(self):
        """
        Get completions in the last 4 days.

        Returns:
            dict: {grid_queue_id: {status, submit_dir} }
        """
        date = (datetime.now()-timedelta(days=4)).isoformat().split('.',1)[0]
        cmd = ['sacct', '-u', getpass.getuser(), '-n', '-P', '-S', date, '-o', 'JobIDRaw,State,JobName,ExitCode,Workdir']
        out = await check_output(*cmd)
        ret = {}
        for line in out.split('\n'):
            if not line.strip():
                continue
            gid,status,name,exit_code,workdir = line.strip().split('|')
            if status != 'COMPLETED' or not name.startswith('iceprod'):
                continue
            if exit_code == '0:0':
                status = 'ok'
            else:
                status = 'error'
            ret[gid] = {'status':status,'submit_dir':workdir}
        return ret

    async def remove(self,tasks):
        """Remove tasks from queueing system."""
        if tasks:
            cmd = ['scancel']+list(tasks)
            await check_call(*cmd)



#!/usr/bin/env python
"""
Server process for starting IceProd.
"""

import asyncio
import os
import sys
import logging
import signal
import importlib
from functools import partial

# Assuming that the directory structure remains constant, this will
# automatically set the python path to find iceprod
bin_dir = os.path.dirname( os.path.abspath(sys.argv[0]) )
root_path = os.path.dirname( bin_dir )
if not root_path in sys.path:
    sys.path.append(root_path)

import iceprod
import iceprod.server
import iceprod.core.logger
from iceprod.server.server import Server

data_class_path = iceprod.server.get_pkgdata_filename('iceprod.server','data/www/dataclasses.js')
if not os.path.exists(data_class_path):
    print('Generating data classes')
    import inspect
    import json
    from iceprod.core import dataclasses
    dcs = {}
    names = dataclasses._plurals.copy()
    for name, obj in inspect.getmembers(dataclasses,inspect.isclass):
        if name[0] != '_' and dict in inspect.getmro(obj):
            dcs[name] = obj().output()
            names[name] = obj.plural
    data = {'classes':dcs,'names':names}
    with open(data_class_path,'w') as f:
        f.write('var dataclasses='+json.dumps(data,separators=(',',':'))+';')

def check_module(name, message='', required=False):
    try:
        importlib.import_module(name)
    except ImportError:
        print('Cannot import python module %s. %s' % (name, message))
        if required:
            print('Required module "%s" not found. Exiting...' % name)
            exit(0)

def check_dependencies():
    check_module('setproctitle', 'Will not be able to set process title.')
    check_module('tornado', required=True)
    check_module('jsonschema')
    check_module('rest_tools', required=True)

check_dependencies()

# start server
async def start_server(*args, **kwargs):
    s = Server(*args, **kwargs)
    s.start()
    try:
        await asyncio.Event().wait()
    finally:
        s.stop()

def runner(stdout=False, *args, **kwargs):
    # set logger
    if stdout:
        iceprod.core.logger.set_logger()
    else:
        iceprod.core.logger.set_logger(logfile='iceprod_server.log')

    # Change name of process for ps
    try:
        from setproctitle import setproctitle
        setproctitle('iceprod_server.main')
    except Exception:
        logging.warning("could not rename process")

    asyncio.run(start_server(*args, **kwargs))
    logging.warning('iceprod exiting')

def setup_I3PROD():
    if 'I3PROD' not in os.environ:
        os.environ['I3PROD'] = os.getcwd()
    for d in ('var/log','var/run','etc'):
        d = os.path.join(os.environ['I3PROD'],d)
        if not os.path.exists(d):
            os.makedirs(d)

def main():
    import argparse

    def key_val(val):
        if '=' not in val:
            msg = f'{val} is not of the form key=val'
            raise argparse.ArgumentTypeError(msg)
        return val

    # Override values with cmdline options
    parser = argparse.ArgumentParser(description='IceProd Server')
    parser.add_argument('-n','--non-daemon',dest='daemon',action='store_false',
                        default=True,help='Do not daemonize')
    parser.add_argument('action',nargs='?',type=str,default='start',
                        choices=['start','stop','kill','hardkill','restart'])
    parser.add_argument('--pidfile',type=str,default='$I3PROD/var/run/iceprod.pid',
                        help='PID lockfile')
    parser.add_argument('--umask',type=int,default=0o077, #octal
                        help='File creation umask')
    parser.add_argument('-c','--config-param',dest='config',action='append',type=key_val,
                        help='extra config params, ex: rest_api.auth_key=XXX')
    args = parser.parse_args()

    setup_I3PROD()

    run = partial(runner, config_params=args.config)

    if args.daemon:
        # now daemonize
        from iceprod.server.daemon import Daemon
        pidfile = os.path.expanduser(os.path.expandvars(args.pidfile))
        chdir = os.path.expanduser(os.path.expandvars('$I3PROD'))
        umask = args.umask
        run = partial(run, outfile=os.path.abspath('var/log/out'), errfile=os.path.abspath('var/log/err'))
        d = Daemon(pidfile, run,
                   chdir=chdir,
                   umask=umask,
                   stdout='var/log/out',
                   stderr='var/log/err')
        getattr(d,args.action)()
    else:
        if args.action != 'start':
            raise Exception('only daemon-mode can execute actions')
        else:
            run(stdout=True)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Cmd line debugger for IceProd
"""

from __future__ import absolute_import, division, print_function

import os
import sys
import signal
import readline
import parser as code_parser
from functools import partial

from IPython.terminal.embed import InteractiveShellEmbed

import iceprod.server.basic_config
import iceprod.server.RPCinternal

def main(cfgfile):
    cfg = iceprod.server.basic_config.BasicConfig()
    cfg.read_file(cfgfile)
    class Response:
        pass
    kwargs = {
        'address':cfg.messaging_url,
        'block':False,
        'service_name':'debug',
        'service_class':Response(),
        'async':False,
    }
    messaging = iceprod.server.RPCinternal.RPCService(**kwargs)
    messaging.start()
    
    try:
        InteractiveShellEmbed(banner1='Launching ipython. The `messaging` object is available.')()
    finally:
        messaging.stop()

if __name__ == '__main__':
    import argparse

    # Override values with cmdline options
    parser = argparse.ArgumentParser(description='IceProd Debugger')
    parser.add_argument('-f','--file',dest='file',type=str,default=None,
                        help='Path to config file')
    args = parser.parse_args()
    
    if args.file:
        cfgfile = os.path.expanduser(os.path.expandvars(args.file))
    else:
        cfgfile = iceprod.server.basic_config.locateconfig()
    
    main(cfgfile)



"""
Clean job temp directories.

Check job temp directories, and if the job is complete then delete it.
"""

import argparse
import asyncio
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from functools import partial
import logging
import os

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.core.gridftp import GridFTP
from iceprod.server.config import IceProdConfig
from iceprod.server.util import str2datetime

logger = logging.getLogger('job_temp_cleaning')


async def run(rest_client, cfg, executor, dataset=None, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        cfg (:py:class:`iceprod.server.config.IceProdConfig`): iceprod config
        debug (bool): debug flag to propagate exceptions
    """
    if 'queue' not in cfg or 'site_temp' not in cfg['queue']:
        return

    suspend_time = timedelta(days=90)
    now = datetime.utcnow()

    try:
        # get all the job_indexes currently in tmp
        temp_dir = cfg['queue']['site_temp']
        dataset_dirs = await asyncio.wrap_future(executor.submit(partial(GridFTP.list, temp_dir, details=True)))
        logger.debug('dataset_dirs: %r', dataset_dirs)
        ret = await rest_client.request('GET', '/datasets?keys=dataset_id|dataset')
        datasets = {}
        for d in ret:
            datasets[str(ret[d]['dataset'])] = ret[d]['dataset_id']
        for entry in dataset_dirs:
            if not entry.directory:
                continue
            d = entry.name
            if dataset and d != dataset:
                continue
            logger.info('temp cleaning for dataset %r', d)
            try:
                job_dirs = await asyncio.wrap_future(executor.submit(partial(GridFTP.list, os.path.join(temp_dir, d), details=True)))
            except Exception:
                logger.error('failed to get job dirs for dataset %r', d, exc_info=True)
                if debug:
                    raise
                continue
            logger.debug('job_dirs: %r', job_dirs)
            try:
                jobs = await rest_client.request('GET', f'/datasets/{datasets[d]}/jobs', {'keys':'status|status_changed|job_index'})
            except Exception:
                logger.error('failed to get jobs for dataset %r', d, exc_info=True)
                if debug:
                    raise
                continue
            logger.debug('jobs: %r', jobs)
            job_indexes = set()
            for job in jobs.values():
                if job['status'] == 'complete' or (job['status'] != 'processing' and now - str2datetime(job['status_changed']) > suspend_time):
                    job_indexes.add(job['job_index'])
            logger.debug('job_indexes: %r', job_indexes)
            for job in job_dirs:
                if not job.directory:
                    continue
                j = job.name
                if not j.isnumeric():
                    if debug:
                        logger.info('j is not numeric: %r', j)
                        raise Exception('not numeric')
                    continue
                if int(j) in job_indexes:
                    try:
                        dagtemp = os.path.join(temp_dir, d, j)
                        logger.info('cleaning site_temp %r', dagtemp)
                        await asyncio.wrap_future(executor.submit(partial(GridFTP.rmtree, dagtemp)))
                    except Exception:
                        logger.warning('failed to clean site_temp', exc_info=True)
                        if debug:
                            raise

    except Exception:
        logger.error('error checking job temp', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('-d', '--dataset', type=str, help='dataset num (optional)')
    parser.add_argument('--site-temp', default='gsiftp://gridftp-scratch.icecube.wisc.edu/mnt/tank/simprod', help='site temp location')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    cfg = IceProdConfig()
    if 'queue' not in cfg:
        cfg['queue'] = {}
    if 'site_temp' not in cfg['queue']:
        cfg['queue']['site_temp'] = args.site_temp

    pool = ThreadPoolExecutor()

    asyncio.run(run(rest_client, cfg, pool, dataset=args.dataset, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Classes supporting dataset serialization.

The interface is as follows::

    class Serialization:
        def dump(config, filename, **kwargs):
            return None
        def dumps(config, **kwargs):
            return string
        def load(filename, **kwargs):
            return config
        def loads(string, **kwargs):
            return config

The kwargs for each function are optional keyword arguments to pass
to the underlying serialization library. Each function is a static method
and can be called like::

    Serialization.dump(config,filename)
"""

import logging
import json as _json

from iceprod.core import dataclasses

logger = logging.getLogger('serialization')


class SerializationError(Exception):
    """An exception that occurs during serialization."""
    def __init__(self, value=''):
        self.value = value

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self.value:
            return 'SerializationError(%r)'%(self.value)
        else:
            return 'SerializationError()'

    def __reduce__(self):
        return (SerializationError,(self.value,))


def dict_to_dataclasses(input_dict):
    """
    Convert a dictionary to dataclasses objects.

    :param input_dict: input dictionary
    :returns: :class:`iceprod.core.dataclasses.Job`
    """
    ret = dataclasses.Job(input_dict)
    ret.convert()
    return ret


class serialize_json(object):
    """
    Serialize a dataset config to json.
    """
    @staticmethod
    def dump(obj, filename, **kwargs):
        return _json.dump(obj, open(filename,'w'), **kwargs)

    @staticmethod
    def dumps(obj, **kwargs):
        return _json.dumps(obj, **kwargs)

    @staticmethod
    def load(filename, **kwargs):
        return dict_to_dataclasses(_json.load(open(filename), **kwargs))

    @staticmethod
    def loads(obj, **kwargs):
        return dict_to_dataclasses(_json.loads(obj, **kwargs))



"""
Update task priority.
"""

import argparse
import asyncio
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.server.priority import Priority

logger = logging.getLogger('update_task_priority')


async def run(rest_client, dataset_id=None, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        dataset_id (str): (optional) dataset id to update
        debug (bool): debug flag to propagate exceptions
    """
    prio = Priority(rest_client)
    try:
        args = {
            'status': 'waiting|queued|processing|reset',
            'keys': 'task_id|depends|dataset_id',
        }
        if dataset_id:
            ret = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks', args)
            tasks = ret.values()
        else:
            ret = await rest_client.request('GET', '/tasks', args)
            tasks = ret['tasks']

        async def check_deps(task):
            dep_futures = []
            for dep in task.get('depends', []):
                t = asyncio.create_task(rest_client.request('GET', f'/tasks/{dep}'))
                dep_futures.append(t)
            for ret in await asyncio.gather(*dep_futures):
                if ret['status'] != 'complete':
                    logger.info('dependency not met for task %s', task['task_id'])
                    return None
            return task

        # check dependencies
        futures = set()
        tasks2 = []
        for task in tasks:
            if len(futures) >= 20:
                done, pending = await asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
                for t in done:
                    ret = await t
                    if ret:
                        tasks2.append(ret)
                futures = pending
            t = asyncio.create_task(check_deps(task))
            futures.add(t)
        while futures:
            done, pending = await asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
            for t in done:
                ret = await t
                if ret:
                    tasks2.append(ret)
            futures = pending
        logger.warning(f'len(tasks) = {len(tasks)}')
        logger.warning(f'len(tasks2) = {len(tasks2)}')

        # update priorities
        futures = set()
        for task in tasks2:
            if len(futures) >= 20:
                done, pending = await asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
                futures = pending
            p = await prio.get_task_prio(task['dataset_id'], task['task_id'])
            logger.info('updating priority for %s.%s = %.4f', task['dataset_id'], task['task_id'], p)
            t = asyncio.create_task(rest_client.request('PATCH', f'/tasks/{task["task_id"]}', {'priority': p}))
            futures.add(t)
        while futures:
            done, pending = await asyncio.wait(futures)
            futures = pending

    except Exception:
        logger.error('error updating task priority', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('-d', '--dataset', type=str, default=None, help='dataset id (optional)')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')
    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, dataset_id=args.dataset, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Interface for configuring and submitting jobs on a computing cluster.
Do not use this class directly. Instead use one of the implementations
that inherit from this class.
"""

import os
import random
import logging
import time
from copy import deepcopy
from datetime import datetime,timedelta
from collections import Counter, defaultdict
import hashlib
import socket
import asyncio
from urllib.parse import urlparse

from asyncache import cached
from cachetools import TTLCache
from cachetools.func import ttl_cache
from tornado.concurrent import run_on_executor

import iceprod
import iceprod.core.exe
from iceprod.core.defaults import add_default_options
from iceprod.core import dataclasses
from iceprod.core import functions
from iceprod.core import serialization
from iceprod.core.resources import Resources, group_hasher
from iceprod.s3 import S3
from iceprod.server import get_pkg_binary


logger = logging.getLogger('grid')


def get_host():
    """Cache the host fqdn for 1 hour"""
    t = time.time()
    if get_host.history and get_host.history[0]+3600 < t:
        return get_host.history[1]
    host = socket.getfqdn()
    get_host.history = [t, host]
    return host
get_host.history = None  # noqa: E305


CRED_SUBMIT_DIR = 'iceprod_credentials'


class BaseGrid(object):
    """
    Interface for a generic job distribution system.
    Do not use this class directly.  Use one of the plugins.
    """

    # use only these grid states when defining grid status
    GRID_STATES = ('queued','processing','completed','error','unknown')

    def __init__(self, gridspec, queue_cfg, cfg, modules, executor, statsd, rest_client, cred_client):
        self.gridspec = gridspec
        self.queue_cfg = queue_cfg
        self.cfg = cfg
        self.modules = modules
        self.executor = executor
        self.statsd = statsd
        self.rest_client = rest_client
        self.cred_client = cred_client

        self.site = None
        if 'site' in self.queue_cfg:
            self.site = self.queue_cfg['site']

        self.credentials_dir = os.path.expanduser(os.path.expandvars(
            self.cfg['queue']['credentials_dir']))
        if not os.path.exists(self.credentials_dir):
            try:
                os.makedirs(self.credentials_dir)
            except Exception:
                logger.warning('error making credentials dir %s', self.credentials_dir,
                               exc_info=True)

        self.submit_dir = os.path.expanduser(os.path.expandvars(
            self.cfg['queue']['submit_dir']))
        if not os.path.exists(self.submit_dir):
            try:
                os.makedirs(self.submit_dir)
            except Exception:
                logger.warning('error making submit dir %s',self.submit_dir,
                               exc_info=True)

        self.grid_processing = 0
        self.grid_idle = 0

    # Public Functions #

    async def check_and_clean(self):
        """Check and clean the grid"""
        # get time limits
        try:
            queued_time = timedelta(seconds=self.queue_cfg['max_task_queued_time'])
        except Exception:
            queued_time = timedelta(seconds=86400*2)
        try:
            processing_time = timedelta(seconds=self.queue_cfg['max_task_processing_time'])
        except Exception:
            processing_time = timedelta(seconds=86400*2)
        try:
            suspend_time = timedelta(seconds=self.queue_cfg['suspend_submit_dir_time'])
        except Exception:
            suspend_time = timedelta(seconds=86400)
        all_time = queued_time + processing_time + suspend_time
        time_dict = {'queued': queued_time,
                     'processing': queued_time+processing_time,
                     'completed': all_time,
                     'error': all_time,
                     'unknown': all_time}
        for t in time_dict:
            logger.debug("time limit: %s - %s",t,time_dict[t])
        now = datetime.utcnow()

        # get pilots from iceprod
        host = get_host()
        args = {
            'queue_host': host,
            'keys': 'pilot_id|queue_host|grid_queue_id|submit_date|tasks',
        }
        ret = await self.rest_client.request('GET', '/pilots', args)

        # filter by queue host
        # index by grid_queue_id
        pilots = {}
        for pilot_id in ret:
            if (ret[pilot_id]['queue_host'] == host
                    and 'grid_queue_id' in ret[pilot_id]
                    and ret[pilot_id]['grid_queue_id']):
                pilots[ret[pilot_id]['grid_queue_id']] = ret[pilot_id]

        # get grid status
        grid_jobs = await asyncio.ensure_future(self.get_grid_status())

        logger.debug("iceprod pilots: %r", list(pilots))
        logger.debug("grid jobs: %r", list(grid_jobs))

        reset_pilots = set(pilots).difference(grid_jobs)
        remove_grid_jobs = set(grid_jobs).difference(pilots)

        prechecked_dirs = set()

        # check the queue
        grid_idle = 0
        for grid_queue_id in set(grid_jobs).intersection(pilots):
            status = grid_jobs[grid_queue_id]['status']
            submit_time = pilots[grid_queue_id]['submit_date']
            if '.' in submit_time:
                submit_time = datetime.strptime(submit_time, '%Y-%m-%dT%H:%M:%S.%f')
            else:
                submit_time = datetime.strptime(submit_time, '%Y-%m-%dT%H:%M:%S')

            if now - submit_time > time_dict[status]:
                logger.info('pilot over time: %r', pilots[grid_queue_id]['pilot_id'])
                reset_pilots.add(pilots[grid_queue_id]['pilot_id'])
            elif status == 'queued':
                grid_idle += 1

            submit_dir = grid_jobs[grid_queue_id]['submit_dir']
            if submit_dir:
                # queueing systems don't like deleteing directories they know
                # about, so put them on a list of "don't touch"
                prechecked_dirs.add(submit_dir)
        self.grid_idle = grid_idle
        self.grid_processing = len(pilots)-len(reset_pilots)-grid_idle

        # check submit directories
        delete_dirs = set()
        for x in os.listdir(self.submit_dir):
            d = os.path.join(self.submit_dir,x)
            if d in prechecked_dirs:
                continue
            if os.path.isdir(d) and '_' in x:
                logger.debug('found submit_dir %s',d)
                mtime = datetime.utcfromtimestamp(os.path.getmtime(d))
                # use all_time instead of suspend_time because the
                # dir will have the submit time, not the last time
                if now-mtime < all_time:
                    continue  # skip for suspended or failed tasks
                delete_dirs.add(d)

        logger.info('%d processing pilots', self.grid_processing)
        logger.info('%d queued pilots', self.grid_idle)
        logger.info('%d ->reset', len(reset_pilots))
        logger.info('%d ->grid remove', len(remove_grid_jobs))
        logger.info('%d ->submit clean', len(delete_dirs))
        self.statsd.gauge('processing_pilots', self.grid_processing)
        self.statsd.gauge('queued_pilots', self.grid_idle)
        self.statsd.incr('reset_pilots', len(reset_pilots))
        self.statsd.incr('grid_remove', len(remove_grid_jobs))
        self.statsd.incr('clean_dirs', len(delete_dirs))

        # reset tasks
        if reset_pilots:
            logger.info('reset %r',reset_pilots)
            for grid_queue_id in reset_pilots:
                try:
                    pilot_id = pilots[grid_queue_id]['pilot_id']
                    await self.rest_client.request('DELETE', '/pilots/{}'.format(pilot_id))
                except KeyError:
                    pass

        # remove grid tasks
        if remove_grid_jobs:
            logger.info('remove %r',remove_grid_jobs)
            await asyncio.ensure_future(self.remove(remove_grid_jobs))

        if delete_dirs:
            await asyncio.ensure_future(self._delete_dirs(delete_dirs))

    async def queue(self):
        """Queue tasks to the grid"""
        # get tasks on the queue
        args = {
            'status': 'queued',
            'keys': 'task_id|dataset_id|status_changed|requirements',
        }
        ret = await self.rest_client.request('GET', '/tasks', args)
        tasks = []
        for t in ret['tasks']:
            if 'site' in t['requirements'] and t['requirements']['site'] != self.site:
                continue
            tasks.append(t)
        dataset_ids = set(row['dataset_id'] for row in tasks)

        # get dataset priorities
        dataset_prios = {}
        for d in dataset_ids:
            dataset = await self.rest_client.request('GET', '/datasets/{}'.format(d))
            dataset_prios[d] = dataset['priority']

        # sort by dataset priority, status changed time
        tasks.sort(key=lambda t:(-1*dataset_prios[t['dataset_id']],t['status_changed'],t['task_id']))

        # queue new pilots
        await self.setup_pilots(tasks)

    # Private Functions #

    def get_queue_num(self, available=100000):
        """Determine how many pilots to queue."""
        tasks_on_queue = self.queue_cfg['pilots_on_queue']
        queue_tot_max = tasks_on_queue[1] - self.grid_processing - self.grid_idle
        queue_idle_max = tasks_on_queue[0] - self.grid_idle
        queue_interval_max = tasks_on_queue[2] if len(tasks_on_queue) > 2 else tasks_on_queue[0]
        queue_num = max(0,min(available - self.grid_idle, queue_tot_max,
                              queue_idle_max, queue_interval_max))
        logger.info('queueing %d pilots', queue_num)
        self.statsd.incr('queueing_pilots', queue_num)
        return queue_num

    @run_on_executor
    def _delete_dirs(self, dirs):
        # delete dirs that need deleting
        for t in dirs:
            if not t.startswith(self.submit_dir):
                # some security against nefarious things
                raise Exception('directory %s not in submit_dir %s'%(t, self.submit_dir))
            try:
                logger.info('deleting submit_dir %s', t)
                functions.removedirs(t)
            except Exception:
                logger.warning('could not delete submit dir %s', t, exc_info=True)
                continue

    @staticmethod
    def _get_resources(tasks):
        """yield resource information for each task in a list"""
        default_resource = deepcopy(Resources.defaults)
        for k in default_resource:
            if isinstance(default_resource[k],list):
                default_resource[k] = len(default_resource[k])
        for t in tasks:
            values = {}
            try:
                for k in t['reqs']:
                    if k in default_resource and t['reqs'][k]:
                        try:
                            if isinstance(default_resource[k], int):
                                values[k] = int(t['reqs'][k])
                            elif isinstance(default_resource[k], float):
                                values[k] = float(t['reqs'][k])
                            else:
                                values[k] = t['reqs'][k]
                        except Exception:
                            logger.warning('bad reqs value for task %r', t)
                    elif k == 'os' and t['reqs'][k]:
                        logger.debug('OS req: %s', t['reqs'][k])
                        values['os'] = ','.join(t['reqs'][k]) if isinstance(t['reqs'][k], list) else t['reqs'][k]
            except TypeError:
                logger.warning('t[reqs]: %r',t['reqs'])
                raise
            resource = deepcopy(default_resource)
            resource.update(values)
            yield resource

    async def setup_pilots(self, tasks):
        """Setup pilots for the task reqs"""
        host = get_host()

        debug = False
        if ('queue' in self.cfg and 'debug' in self.cfg['queue']
                and self.cfg['queue']['debug']):
            debug = True

        # convert to resource requests and group them
        groups = defaultdict(list)
        for resources in self._get_resources({'reqs':t['requirements']} for t in tasks):
            k = group_hasher(resources)
            groups[k].append(resources)

        # get already queued pilots
        pilot_groups = Counter()
        ret = await self.rest_client.request('GET', '/pilots', {'host': '', 'keys':'resources'})
        for pilot in ret.values():
            k = group_hasher(pilot['resources'])
            pilot_groups[k] += 1

        # remove already queued groups from consideration
        groups_considered = Counter()
        for k in groups:
            n = len(groups[k]) - pilot_groups[k]
            if n > 0:
                groups_considered[k] = n

        # select at least one from each resource group
        queue_num = self.get_queue_num(available=len(tasks))
        groups_to_queue = Counter()
        keys = set(groups_considered.keys())
        while queue_num > 0 and keys:
            for k in list(keys):
                if groups_considered[k] < 1:
                    keys.remove(k)
                else:
                    groups_to_queue[k] += 1
                    groups_considered[k] -= 1
                    queue_num -= 1
                    if queue_num < 1:
                        break
        logger.debug('groups_to_queue: %r', groups_to_queue)

        for r in groups_to_queue:
            try:
                resources = defaultdict(list)
                for x in groups[r]:
                    for k in x:
                        if x[k] is not None:
                            resources[k].append(x[k])
                resources = {k:resources[k][0] if isinstance(resources[k][0],dataclasses.String) else max(resources[k])
                             for k in resources}
                logger.info('submitting %d pilots for resource %r',
                            groups_to_queue[r], resources)
                for name in resources:
                    self.statsd.incr('pilot_resources.'+name, resources[name])
                pilot = {
                    'task_id': 'pilot',
                    'name': 'pilot',
                    'debug': debug,
                    'reqs': resources,
                    'num': groups_to_queue[r],
                    'pilot_ids': [],
                }

                args = {
                    'queue_host': host,
                    'queue_version': iceprod.__version__,
                    'resources': resources,
                }
                for _ in range(groups_to_queue[r]):
                    ret = await self.rest_client.request('POST', '/pilots', args)
                    pilot['pilot_ids'].append(ret['result'])

                await self.setup_submit_directory(pilot)
                await asyncio.ensure_future(self.submit(pilot))

                grid_queue_ids = pilot['grid_queue_id'].split(',')
                for i,pilot_id in enumerate(pilot['pilot_ids']):
                    ret = await self.rest_client.request(
                        'PATCH',
                        '/pilots/{}'.format(pilot_id),
                        {'grid_queue_id': grid_queue_ids[i]}
                    )

            except Exception:
                logger.error('error submitting pilots', exc_info=True)

    @ttl_cache(ttl=600)
    def get_token(self):
        return self.cred_client.make_access_token()

    @cached(TTLCache(1024, 60))
    async def get_user_credentials(self, username):
        ret = await self.cred_client.request('GET', f'/users/{username}/credentials')
        return ret

    @cached(TTLCache(1024, 60))
    async def get_group_credentials(self, group):
        ret = await self.cred_client.request('GET', f'/groups/{group}/credentials')
        return ret

    async def customize_task_config(self, task_cfg, job_cfg=None, dataset=None):
        """Transforms for the task config"""
        logger.info('customize_task_config for %s', job_cfg.get('options', {}).get('task_id', 'unknown'))

        # first expand site temp urls
        def expand_remote(cfg):
            new_data = []
            for d in cfg.get('data', []):
                if not d['remote']:
                    try:
                        remote_base = d.storage_location(job_cfg)
                        logger.info('expanding remote for %r', d['local'])
                        d['remote'] = os.path.join(remote_base, d['local'])
                    except Exception:
                        # ignore failed expansions, as these are likely local temp paths
                        pass
                new_data.append(d)
            cfg['data'] = new_data
        expand_remote(task_cfg)
        for tray in task_cfg['trays']:
            expand_remote(tray)
            for module in tray['modules']:
                expand_remote(module)
        logger.info('task_cfg: %r', task_cfg)

        # now apply S3 and token credentials
        creds = self.cfg.get('creds', {})
        if dataset['group'] == 'users':
            ret = await self.get_user_credentials(dataset['username'])
        else:
            ret = await self.get_group_credentials(dataset['group'])
        creds.update(ret)

        s3_creds = {url: creds.pop(url) for url in list(creds) if creds[url]['type'] == 's3'}
        if s3_creds:
            # if we have any s3 credentials, try presigning urls
            logger.info('testing job for s3 credentials')
            try:
                queued_time = timedelta(seconds=self.queue_cfg['max_task_queued_time'])
            except Exception:
                queued_time = timedelta(seconds=86400*2)
            try:
                processing_time = timedelta(seconds=self.queue_cfg['max_task_processing_time'])
            except Exception:
                processing_time = timedelta(seconds=86400*2)
            expiration = (queued_time + processing_time).seconds

            def presign_s3(cfg):
                new_data = []
                for d in cfg.get('data', []):
                    for url in s3_creds:
                        if d['remote'].startswith(url):
                            logger.info('found data for cred: %s', url)
                            path = d['remote'][len(url):].lstrip('/')
                            bucket = None
                            if '/' in path:
                                bucket, key = path.split('/', 1)
                            if (not bucket) or bucket not in s3_creds[url]['buckets']:
                                key = path
                                bucket = urlparse(url).hostname.split('.', 1)[0]
                                if bucket not in s3_creds[url]['buckets']:
                                    raise RuntimeError('bad s3 bucket')

                            s = S3(url, s3_creds[url]['access_key'], s3_creds[url]['secret_key'], bucket=bucket)
                            if d['movement'] == 'input':
                                d['remote'] = s.get_presigned(key, expiration=expiration)
                                new_data.append(d)
                            elif d['movement'] == 'output':
                                d['remote'] = s.put_presigned(key, expiration=expiration)
                            elif d['movement'] == 'both':
                                d['movement'] = 'input'
                                d['remote'] = s.get_presigned(key, expiration=expiration)
                                new_data.append(d.copy())
                                d['movement'] = 'output'
                                d['remote'] = s.put_presigned(key, expiration=expiration)
                            else:
                                raise RuntimeError('unknown s3 data movement')
                            new_data.append(d)
                            break
                    else:
                        new_data.append(d)
                cfg['data'] = new_data

            presign_s3(task_cfg)
            for tray in task_cfg['trays']:
                presign_s3(tray)
                for module in tray['modules']:
                    presign_s3(module)

        oauth_creds = {url: creds.pop(url) for url in list(creds) if creds[url]['type'] == 'oauth'}
        if oauth_creds:
            # if we have token-based credentials, add them to the config
            logger.info('testing job for oauth credentials')
            cred_keys = set()

            def get_creds(cfg):
                for d in cfg.get('data', []):
                    for url in oauth_creds:
                        if d['remote'].startswith(url):
                            logger.info('found data for cred: %s', url)
                            cred_keys.add(url)
                            break

            get_creds(task_cfg)
            for tray in task_cfg['trays']:
                get_creds(tray)
                for module in tray['modules']:
                    get_creds(module)

            file_creds = {}
            for url in cred_keys:
                cred_name = hashlib.sha1(oauth_creds[url]['access_token'].encode('utf-8')).hexdigest()
                path = os.path.join(self.credentials_dir, cred_name)
                if not os.path.exists(path):
                    with open(path, 'w') as f:
                        f.write(oauth_creds[url]['access_token'])
                file_creds[url] = cred_name
            job_cfg['options']['credentials'] = file_creds

    async def setup_submit_directory(self,task):
        """Set up submit directory"""
        # create directory for task
        submit_dir = self.submit_dir
        task_dir = os.path.join(submit_dir,task['task_id']+'_'+str(random.randint(0,1000000)))
        while os.path.exists(task_dir):
            task_dir = os.path.join(submit_dir,task['task_id']+'_'+str(random.randint(0,1000000)))
        task_dir = os.path.abspath(os.path.expanduser(os.path.expandvars(task_dir)))
        os.makedirs(task_dir)
        task['submit_dir'] = task_dir

        # symlink or copy the .sh file
        src = get_pkg_binary('iceprod', 'loader.sh')
        dest = os.path.join(task_dir, 'loader.sh')
        try:
            os.symlink(src, dest)
        except Exception:
            try:
                functions.copy(src, dest)
            except Exception:
                logger.error('Error creating symlink or copy of .sh file: %s',dest,exc_info=True)
                raise

        # get passkey
        # expiration = self.queue_cfg['max_task_queued_time']
        # expiration += self.queue_cfg['max_task_processing_time']
        # expiration += self.queue_cfg['max_task_reset_time']
        # TODO: take expiration into account
        passkey = self.get_token()

        # write cfg
        cfg, filelist = self.write_cfg(task)

        # create submit file
        try:
            await asyncio.ensure_future(self.generate_submit_file(
                task,
                cfg=cfg,
                passkey=passkey,
                filelist=filelist
            ))
        except Exception:
            logger.error('Error generating submit file',exc_info=True)
            raise

    def create_config(self, task):
        if 'config' in task and task['config']:
            config = serialization.dict_to_dataclasses(task['config'])
        else:
            config = dataclasses.Job()

        # add server options
        config['options']['task_id'] = task['task_id']
        config['options']['task'] = task['name']
        if 'job' in task:
            config['options']['job'] = task['job']
        if 'jobs_submitted' in task:
            config['options']['jobs_submitted'] = task['jobs_submitted']
        if 'dataset_id' in task:
            config['options']['dataset_id'] = task['dataset_id']
        if 'dataset' in task:
            config['options']['dataset'] = task['dataset']
        config['options']['stillrunninginterval'] = self.queue_cfg['ping_interval']
        config['options']['debug'] = task['debug']
        config['options']['upload'] = 'logging'
        config['options']['gridspec'] = self.gridspec
        if 'site_temp' in self.cfg['queue']:
            config['options']['site_temp'] = self.cfg['queue']['site_temp']
        if ('download' in self.cfg and 'http_username' in self.cfg['download']
                and self.cfg['download']['http_username']):
            config['options']['username'] = self.cfg['download']['http_username']
        if ('download' in self.cfg and 'http_password' in self.cfg['download']
                and self.cfg['download']['http_password']):
            config['options']['password'] = self.cfg['download']['http_password']

        add_default_options(config['options'])

        return config

    def write_cfg(self, task):
        """Write the config file for a task-like object"""
        filename = os.path.join(task['submit_dir'],'task.cfg')
        filelist = [filename]

        config = self.create_config(task)
        if creds := config['options'].get('credentials', {}):
            cred_dir = os.path.join(task['submit_dir'], CRED_SUBMIT_DIR)
            os.mkdir(cred_dir)
            for name in creds.values():
                src = os.path.join(self.credentials_dir, name)
                dest = os.path.join(cred_dir, name)
                os.symlink(src, dest)
            filelist.append(cred_dir)
        if 'system' in self.cfg and 'remote_cacert' in self.cfg['system']:
            config['options']['ssl'] = {}
            config['options']['ssl']['cacert'] = os.path.basename(self.cfg['system']['remote_cacert'])
            src = self.cfg['system']['remote_cacert']
            dest = os.path.join(task['submit_dir'],config['options']['ssl']['cacert'])
            try:
                os.symlink(src,dest)
            except Exception:
                try:
                    functions.copy(src,dest)
                except Exception:
                    logger.error('Error creating symlink or copy of remote_cacert',
                                 exc_info=True)
                    raise
            filelist.append(dest)
        if 'x509proxy' in self.cfg['queue'] and self.cfg['queue']['x509proxy']:
            config['options']['x509'] = os.path.basename(self.cfg['queue']['x509proxy'])
            src = self.cfg['queue']['x509proxy']
            logger.info('submit_dir %r  x509 %r', task['submit_dir'], config['options']['x509'])
            dest = os.path.join(task['submit_dir'],config['options']['x509'])
            try:
                os.symlink(src,dest)
            except Exception:
                try:
                    functions.copy(src,dest)
                except Exception:
                    logger.error('Error creating symlink or copy of x509 proxy',
                                 exc_info=True)
                    raise
            filelist.append(dest)
        if 'extra_file_tranfers' in self.cfg['queue'] and self.cfg['queue']['extra_file_tranfers']:
            for f in self.cfg['queue']['extra_file_tranfers']:
                logger.info('submit_dir %r  extra_files %r', task['submit_dir'], f)
                dest = os.path.join(task['submit_dir'],os.path.basename(f))
                try:
                    os.symlink(os.path.abspath(f),dest)
                except Exception:
                    try:
                        functions.copy(f,dest)
                    except Exception:
                        logger.error('Error creating symlink or copy of extra file %s',
                                     f, exc_info=True)
                        raise
                filelist.append(dest)
        if 'data_movement_stats' in self.cfg['queue'] and self.cfg['queue']['data_movement_stats']:
            config['options']['data_movement_stats'] = self.cfg['queue']['data_movement_stats']
        if 'upload_checksum' in self.cfg['queue']:
            config['options']['upload_checksum'] = self.cfg['queue']['upload_checksum']

        if 'reqs' in task:
            # add resources
            config['options']['resources'] = {}
            for r in task['reqs']:
                config['options']['resources'][r] = task['reqs'][r]

        # write to file
        serialization.serialize_json.dump(config, filename)

        c = iceprod.core.exe.Config(config)
        config = c.parseObject(config, {})

        return (config, filelist)

    # not async: called from executor
    def get_submit_args(self,task,cfg=None,passkey=None):
        """Get the submit arguments to start the loader script."""
        # get website address
        if ('rest_api' in self.cfg and self.cfg['rest_api'] and
                'url' in self.cfg['rest_api'] and self.cfg['rest_api']['url']):
            web_address = self.cfg['rest_api']['url']
        else:
            raise Exception('no web address for rest calls')

        args = []
        if 'software_dir' in self.queue_cfg and self.queue_cfg['software_dir']:
            args.append('-s {}'.format(self.queue_cfg['software_dir']))
        if 'iceprod_dir' in self.queue_cfg and self.queue_cfg['iceprod_dir']:
            args.append('-e {}'.format(self.queue_cfg['iceprod_dir']))
        if 'x509proxy' in self.cfg['queue'] and self.cfg['queue']['x509proxy']:
            args.append('-x {}'.format(os.path.basename(self.cfg['queue']['x509proxy'])))
        if ('download' in self.cfg and 'http_proxy' in self.cfg['download']
                and self.cfg['download']['http_proxy']):
            args.apend('-c {}'.format(self.cfg['download']['http_proxy']))
        args.append('--url {}'.format(web_address))
        if passkey:
            args.append('--passkey {}'.format(passkey))
        if cfg:
            args.append('--cfgfile task.cfg')
        if 'debug' in task and task['debug']:
            args.append('--debug')
        return args

    # Plugin Overrides #

    async def get_grid_status(self):
        """Get all tasks running on the queue system.
           Returns {grid_queue_id:{status,submit_dir}}
        """
        return {}

    async def generate_submit_file(self,task,cfg=None,passkey=None,filelist=None):
        """Generate queueing system submit file for task in dir."""
        raise NotImplementedError()

    async def submit(self,task):
        """Submit task to queueing system."""
        raise NotImplementedError()

    async def remove(self,tasks):
        """Remove tasks from queueing system."""
        pass



"""
Tools to help manage Globus proxies
"""

import os
import subprocess
import logging

from iceprod.server.config import IceProdConfig

logger = logging.getLogger('globus')


class SiteGlobusProxy(object):
    """
    Manage site-wide globus proxy

    :param cfgfile: cfgfile location (optional)
    :param duration: proxy duration (optional, default 72 hours)
    """
    def __init__(self, cfgfile=None, duration=None):
        if not cfgfile:
            cfgfile = os.path.join(os.getcwd(),'globus_proxy.json')
        self.cfg = IceProdConfig(filename=cfgfile, defaults=False,
                                 validate=False)
        if duration:
            self.cfg['duration'] = duration
        elif 'duration' not in self.cfg:
            self.cfg['duration'] = 72

    def set_passphrase(self, p):
        """Set the passphrase"""
        self.cfg['passphrase'] = p

    def set_duration(self, d):
        """Set the duration"""
        self.cfg['duration'] = d

    def set_voms_vo(self, vo):
        """Set the voms VO"""
        self.cfg['voms_vo'] = vo

    def set_voms_role(self, r):
        """Set the voms role"""
        self.cfg['voms_role'] = r

    def update_proxy(self):
        """Update the proxy"""
        if 'passphrase' not in self.cfg:
            raise Exception('passphrase missing')
        if 'duration' not in self.cfg:
            raise Exception('duration missing')
        logger.info('duration: %r',self.cfg['duration'])
        if subprocess.call(
            [
                'grid-proxy-info',
                '-e',
                '-valid',
                '%d:0'%self.cfg['duration'],
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        ):
            # proxy needs updating
            if 'voms_vo' in self.cfg and self.cfg['voms_vo']:
                cmd = ['voms-proxy-init']
                if 'voms_role' in self.cfg and self.cfg['voms_role']:
                    vo = self.cfg['voms_vo']
                    role = self.cfg['voms_role']
                    cmd.extend(['-voms', '{0}:/{0}/Role={1}'.format(vo, role)])
                else:
                    cmd.extend(['-voms', self.cfg['voms_vo']])
            else:
                cmd = ['grid-proxy-init']
            cmd.extend(['-pwstdin','-valid','%d:0'%(self.cfg['duration']+1)])
            if 'out' in self.cfg:
                cmd.extend(['-out', self.cfg['out']])
            inputbytes = (self.cfg['passphrase']+'\n').encode('utf-8')
            p = subprocess.run(cmd, input=inputbytes, capture_output=True, timeout=60, check=False)
            logger.info('proxy cmd: %r', p.args)
            logger.info('stdout: %s', p.stdout)
            logger.info('stderr: %s', p.stderr)
            if 'voms_vo' in self.cfg and self.cfg['voms_vo']:
                for line in p.stdout.decode('utf-8').split('\n'):
                    if line.startswith('Creating proxy') and line.endswith('Done'):
                        break  # this is a good proxy
                else:
                    raise Exception('voms-proxy-init failed')
            elif p.returncode > 0:
                raise Exception('grid-proxy-init failed')

    def get_proxy(self):
        """Get the proxy location"""
        if 'out' in self.cfg:
            return self.cfg['out']
        FNULL = open(os.devnull, 'w')
        return subprocess.check_output(['grid-proxy-info','-path'],
                                       stderr=FNULL).decode('utf-8').strip()



"""
Functions used to calculate dataset priority.
"""

import math
import logging
from functools import partial
from collections import defaultdict

from iceprod.server import GlobalID


def apply_group_prios(datasets, groups=None, filters=None):
    """
    Apply the group priorities to the datasets.

    :param datasets: dataset dict with 'priority' and 'groups_id'
    :param groups: dump of groups table
    :param filters: any filters for the groups
    """
    if not groups:
        return datasets

    # first, calculate group priorities with filtering
    base_groups = {groups[g]['name']:groups[g]['priority'] for g in groups}
    if filters:
        filtered_groups = [g for g in base_groups if any(True for f in filters if f in g)]
    else:
        filtered_groups = base_groups.keys()

    group_prios = {}
    for g in filtered_groups:
        parts = g.split('/')
        prio = 1.0
        for i in range(2,len(parts)+1):
            print(i, '/'.join(parts[:i]))
            prio *= base_groups['/'.join(parts[:i])]
        subtract = 0.0
        for k in base_groups:
            if k.startswith(g) and len(k.split('/')) == len(parts)+1:
                subtract += base_groups[k]
        group_prios[g] = prio * (1.0 - subtract)

    # normalize dataset-group priorities
    norm = defaultdict(0.)
    for d in datasets:
        norm[datasets[d]['groups_id']] += datasets[g]['priority']

    # apply group priorities to datasets
    ret = {}
    for d in datasets:
        gid = datasets[d]['groups_id']
        if gid in groups:
            p_g = group_prios[groups[gid]['name']]
            if p_g > 0:
                ret[d] = datasets[d].copy()
                p_d = ret[d]['priority']
                if p_d <= 0:
                    ret[d]['priority'] = p_g
                else:
                    ret[d]['priority'] = p_d*1.0/norm[gid] * p_g
    return ret


def calc_dataset_prio(dataset, queueing_factor_priority=1.0,
                      queueing_factor_dataset=1.0, queueing_factor_tasks=1.0):
    """
    Calculate the dataset priority.

    :param dataset: a dataset with 'dataset_id', 'priority' and 'tasks_submitted'
    :param queueing_factor_priority: queueing factor for priority
    :param queueing_factor_dataset: queueing factor for dataset id
    :param queueing_factor_tasks: queueing factor for number of tasks
    """
    import math
    # priority factors
    qf_p = queueing_factor_priority
    qf_d = queueing_factor_dataset
    qf_t = queueing_factor_tasks

    # get dataset info
    p = dataset['priority']
    if p < 0 or p > 100:
        # do not allow negative or overly large priorities (they skew things)
        p = 0
        logging.warning('Priority for dataset %s is invalid, using default',dataset['dataset_id'])
    d = GlobalID.localID_ret(dataset['dataset_id'],type='int')
    if d < 0:
        d = 0
        logging.warning('Dataset for dataset %s is invalid, using default',dataset['dataset_id'])
    t = dataset['tasks_submitted']

    # return prio
    if t < 1:
        prio = (qf_p/10.0*p-qf_d/10000.0*d)
    else:
        prio = (qf_p/10.0*p-qf_d/10000.0*d-qf_t/10.0*math.log10(t))
    if prio < 0:
        prio = 0
        logging.error('Dataset prio for dataset %s is <0',dataset['dataset_id'])
    return prio


def calc_datasets_prios(datasets, queueing_factor_priority=1.0,
                        queueing_factor_dataset=1.0, queueing_factor_tasks=1.0):
    """
    Calculate the dataset priority for each dataset, normalized.

    :param datasets: dataset dict with 'dataset_id', 'priority' and 'tasks_submitted'
    :param queueing_factor_priority: queueing factor for priority
    :param queueing_factor_dataset: queueing factor for dataset id
    :param queueing_factor_tasks: queueing factor for number of tasks
    """
    calc = partial(calc_dataset_prio,
                   queueing_factor_priority=queueing_factor_priority,
                   queueing_factor_dataset=queueing_factor_dataset,
                   queueing_factor_tasks=queueing_factor_tasks)

    dataset_prios = {}
    for id in datasets:
        dataset_prios[id] = calc(datasets[id])
    logging.debug('dataset prios: %r',dataset_prios)
    # normalize
    total_prio = math.fsum(dataset_prios.values())
    if total_prio <= 0:
        # datasets do not have priority, so assign all equally
        for d in dataset_prios:
            dataset_prios[d] = 1.0/len(dataset_prios)
    else:
        for d in dataset_prios:
            dataset_prios[d] /= total_prio

    return dataset_prios
